if not IsLoggedIn then
	-- wow 2.0
	return
end
local sm = AceLibrary("SharedMedia-1.0")
local L = AceLibrary("AceLocale-2.2"):new("Quartz")

local QuartzBuff = Quartz:NewModule('Buff')
local self = QuartzBuff
local Quartz = Quartz

local GetTime = GetTime
local table_sort = table.sort
local math_ceil = math.ceil

local db

local new, del
do
	local cache = setmetatable({}, {__mode='k'})
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
end
local OnUpdate
do
	local min = L["%dm"]
	local function timenum(num)
		if num <= 10 then
			return ('%.1f'):format(num)
		elseif num <= 60 then
			return ('%d'):format(num)
		else
			return min:format(math_ceil(num / 60))
		end
	end
	function OnUpdate(frame)
		local currentTime = GetTime()
		local endTime = frame.endTime
		if currentTime > endTime then
			self:UpdateBars()
		else
			local remaining = (currentTime - frame.startTime)
			frame:SetValue(endTime - remaining)
			frame.timetext:SetText(timenum(endTime - currentTime))
		end
		
	end
end
local function OnShow(frame)
	frame:SetScript('OnUpdate', OnUpdate)
end
local function OnHide(frame)
	frame:SetScript('OnUpdate', nil)
end

local framefactory = {
	__index = function(t,k)
		local bar = CreateFrame('StatusBar', nil, UIParent)
		t[k] = bar
		bar:SetFrameStrata('MEDIUM')
		bar:Hide()
		bar:SetScript('OnShow', OnShow)
		bar:SetScript('OnHide', OnHide)
		bar:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16})
		bar:SetBackdropColor(0,0,0)
		bar.text = bar:CreateFontString(nil, 'OVERLAY')
		bar.timetext = bar:CreateFontString(nil, 'OVERLAY')
		bar.icon = bar:CreateTexture(nil, 'DIALOG')
		
		self:ApplySettings()
		return bar
	end
}
local targetbars = setmetatable({}, framefactory)
local focusbars = setmetatable({}, framefactory)

function QuartzBuff:OnInitialize()
	db = Quartz:AcquireDBNamespace("Buff")
	Quartz:RegisterDefaults("Buff", "profile", {
		target = true,
		targetbuffs = true,
		targetdebuffs = true,
		targetposition = L["Top Right"],
		targeticons = true,
		targeticonside = L["Right"],
		
		focus = true,
		focusbuffs = true,
		focusdebuffs = true,
		focusposition = L["Bottom Left"],
		focusicons = true,
		focusiconside = L["Left"],
		
		buffnametext = true,
		bufftimetext = true,
		
		bufftexture = 'LiteStep',
		buffwidth = 120,
		buffheight = 12,
		bufffont = 'Friz Quadrata TT',
		bufffontsize = 9,
		buffalpha = 1,
		
		buffcolor = {0,0.49, 1},
		debuffcolor = {1.0,0.7, 0},
		
		bufftextcolor = {1,1,1},
	})
end
function QuartzBuff:OnEnable()
	self:RegisterBucketEvent("UNIT_AURA", 0.5)
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateBars")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED", "UpdateBars")
	Quartz.ApplySettings()
end
function QuartzBuff:OnDisable()
	for _, v in pairs(targetbars) do
		v:Hide()
	end
	for _, v in pairs(focusbars) do
		v:Hide()
	end
end
function QuartzBuff:UNIT_AURA(units)
	for unit in pairs(units) do
		if unit == 'target' then
			self:UpdateTargetBars()
		end
		if unit == 'focus' or UnitIsUnit('focus', unit) then
			self:UpdateFocusBars()
		end
	end
end
function QuartzBuff:UpdateBars()
	self:UpdateTargetBars()
	self:UpdateFocusBars()
end
do
	local function sort(a,b)
		if a.isbuff == b.isbuff then
			return a.name < b.name
		else
			return a.isbuff
		end
	end
	local tmp = {}
	local called = false -- prevent recursive calls when new bars are created.
	function QuartzBuff:UpdateTargetBars()
		if called then
			return
		end
		called = true
		local db = db.profile
		if db.target then
			local currentTime = GetTime()
			for k in pairs(tmp) do
				tmp[k] = del(tmp[k])
			end
			if db.targetbuffs then
				for i = 1, 32 do
					local name, _, texture, _, duration, remaining = UnitBuff('target', i)
					if not name then
						break
					end
					if duration and duration > 0 then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.isbuff = true
					end
				end
			end
			if db.targetdebuffs then
				for i = 1, 40 do
					local name, _, texture, _, _, duration, remaining = UnitDebuff('target', i)
					if not name then
						break
					end
					if duration and duration > 0 then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
					end
				end
			end
			table_sort(tmp, sort)
			local maxindex = 0
			for k,v in ipairs(tmp) do
				maxindex = k
				local bar = targetbars[k]
				
				bar.text:SetText(v.name)
				bar.icon:SetTexture(v.texture)
				local elapsed = (v.duration - v.remaining)
				local startTime, endTime = (currentTime - elapsed), (currentTime + v.remaining)
				bar.startTime = startTime
				bar.endTime = endTime
				bar:SetMinMaxValues(startTime, endTime)
				bar:Show()
				if v.isbuff then
					bar:SetStatusBarColor(unpack(db.buffcolor))
				else
					bar:SetStatusBarColor(unpack(db.debuffcolor))
				end
			end
			for i = maxindex+1, #targetbars do
				targetbars[i]:Hide()
			end
		else
			for _, v in ipairs(targetbars) do
				v:Hide()
			end
		end
		called = false
	end
	function QuartzBuff:UpdateFocusBars()
		if called then
			return
		end
		called = true
		local db = db.profile
		if db.focus then
			local currentTime = GetTime()
			for k in pairs(tmp) do
				tmp[k] = del(tmp[k])
			end
			if db.focusbuffs then
				for i = 1, 32 do
					local name, _, texture, _, duration, remaining = UnitBuff('focus', i)
					if not name then
						break
					end
					if duration and duration > 0 then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
						t.isbuff = true
					end
				end
			end
			if db.focusdebuffs then
				for i = 1, 40 do
					local name, _, texture, _, _, duration, remaining = UnitDebuff('focus', i)
					if not name then
						break
					end
					if duration and duration > 0 then
						local t = new()
						tmp[#tmp+1] = t
						t.name = name
						t.texture = texture
						t.duration = duration
						t.remaining = remaining
					end
				end
			end
			table_sort(tmp, sort)
			local maxindex = 0
			for k,v in ipairs(tmp) do
				maxindex = k
				local bar = focusbars[k]
				bar.text:SetText(v.name)
				bar.icon:SetTexture(v.texture)
				local elapsed = (v.duration - v.remaining)
				local startTime, endTime = (currentTime - elapsed), (currentTime + v.remaining)
				bar.startTime = startTime
				bar.endTime = endTime
				bar:SetMinMaxValues(startTime, endTime)
				bar:Show()
				if v.isbuff then
					bar:SetStatusBarColor(unpack(db.buffcolor))
				else
					bar:SetStatusBarColor(unpack(db.debuffcolor))
				end
			end
			for i = maxindex+1, #focusbars do
				focusbars[i]:Hide()
			end
		else
			for _, v in ipairs(focusbars) do
				v:Hide()
			end
		end
		called = false
	end
end
do
	local function apply(unit, i, bar, db, direction)
		local bars, position, icons, iconside
		if unit == 'target' then
			bars = targetbars
			position = db.targetposition
			icons = db.targeticons
			iconside = db.targeticonside
		else
			bars = focusbars
			position = db.focusposition
			icons = db.focusicons
			iconside = db.focusiconside
		end
		bar:ClearAllPoints()
		bar:SetStatusBarTexture(sm:Fetch('statusbar', db.bufftexture))
		bar:SetWidth(db.buffwidth)
		bar:SetHeight(db.buffheight)
		bar:SetScale(Quartz.db.profile.scale)
		bar:SetAlpha(db.buffalpha)
		if i == 1 then
			if position == L["Top"] then
				direction = 1
				bar:SetPoint('BOTTOM', QuartzCastBar, 'TOP', 0, 1)
			elseif position == L["Bottom"] then
				direction = -1
				bar:SetPoint('TOP', QuartzCastBar, 'BOTTOM', 0, -1)
			elseif position == L["Left (grow up)"] then
				direction = 1
				bar:SetPoint('BOTTOMRIGHT', QuartzCastBar, 'BOTTOMLEFT', -3, 1)
			elseif position == L["Left (grow down)"] then
				direction = -1
				bar:SetPoint('TOPRIGHT', QuartzCastBar, 'TOPLEFT', -3, -1)
			elseif position == L["Right (grow up)"] then
				direction = 1
				bar:SetPoint('BOTTOMLEFT', QuartzCastBar, 'BOTTOMRIGHT', 3, 1)
			elseif position == L["Right (grow down)"] then
				direction = -1
				bar:SetPoint('TOPLEFT', QuartzCastBar, 'TOPRIGHT', 3, -1)
			elseif position == L["Top Right"] then
				direction = 1
				bar:SetPoint('BOTTOMRIGHT', QuartzCastBar, 'TOPRIGHT', -3, 1)
			elseif position == L["Bottom Right"] then
				direction = -1
				bar:SetPoint('TOPRIGHT', QuartzCastBar, 'BOTTOMRIGHT', -3, -1)
			elseif position == L["Top Left"] then
				direction = 1
				bar:SetPoint('BOTTOMLEFT', QuartzCastBar, 'TOPLEFT', 3, 1)
			else-- L["Bottom Left"]
				direction = -1
				bar:SetPoint('TOPLEFT', QuartzCastBar, 'BOTTOMLEFT', 3, -1)
			end
		else
			if direction == 1 then
				bar:SetPoint('BOTTOMRIGHT', bars[i-1], 'TOPRIGHT', 0, 1)
			else -- -1
				bar:SetPoint('TOPRIGHT', bars[i-1], 'BOTTOMRIGHT', 0, -1)
			end
		end
		
		local text = bar.text
		if db.buffnametext then
			text:Show()
			text:ClearAllPoints()
			text:SetFont(sm:Fetch('font', db.bufffont), db.bufffontsize)
			text:SetShadowColor( 0, 0, 0, 1)
			text:SetShadowOffset( 0.8, -0.8 )
			text:SetTextColor(unpack(db.bufftextcolor))
			text:SetNonSpaceWrap(false)
			text:ClearAllPoints()
			text:SetWidth(db.buffwidth)
			text:SetPoint("LEFT", bar, "LEFT", 2, 0)
			text:SetJustifyH("LEFT")
		else
			text:Hide()
		end
		
		local timetext = bar.timetext
		if db.bufftimetext then
			timetext:Show()
			timetext:ClearAllPoints()
			timetext:SetFont(sm:Fetch('font', db.bufffont), db.bufffontsize)
			timetext:SetShadowColor( 0, 0, 0, 1)
			timetext:SetShadowOffset( 0.8, -0.8 )
			timetext:SetTextColor(unpack(db.bufftextcolor))
			timetext:SetNonSpaceWrap(false)
			timetext:ClearAllPoints()
			timetext:SetWidth(db.buffwidth)
			timetext:SetPoint("RIGHT", bar, "RIGHT", -2, 0)
			timetext:SetJustifyH("RIGHT")
		else
			timetext:Hide()
		end
		
		local icon = bar.icon
		if icons then
			icon:Show()
			icon:SetWidth(db.buffheight-1)
			icon:SetHeight(db.buffheight-1)
			icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
			icon:ClearAllPoints()
			if iconside == L["Left"] then
				icon:SetPoint('RIGHT', bar, "LEFT", -1, 0)
			else
				icon:SetPoint('LEFT', bar, "RIGHT", 1, 0)
			end
		else
			icon:Hide()
		end
		
		return direction
	end
	function QuartzBuff:ApplySettings()
		if Quartz:IsModuleActive('Buff') then
			local db = db.profile
			local direction
			for i, v in pairs(targetbars) do
				direction = apply('target', i, v, db, direction)
			end
			direction = nil
			for i, v in pairs(focusbars) do
				direction = apply('focus', i, v, db, direction)
			end
			self:UpdateBars()
		end
	end
end
do
	local function set(field, value)
		db.profile[field] = value
		Quartz.ApplySettings()
	end
	local function get(field)
		return db.profile[field]
	end
	local function setcolor(field, ...)
		db.profile[field] = {...}
		Quartz.ApplySettings()
	end
	local function getcolor(field)
		return unpack(db.profile[field])
	end
	local positions = {
		L["Bottom"],
		L["Top"],
		L["Top Left"],
		L["Top Right"],
		L["Bottom Left"],
		L["Bottom Right"],
		L["Left (grow up)"],
		L["Left (grow down)"],
		L["Right (grow up)"],
		L["Right (grow down)"],
	}
	Quartz.options.args.buffs = {
		type = 'group',
		name = L["Buffs"],
		desc = L["Buffs"],
		order = 602,
		args = {
			toggle = {
				type = 'toggle',
				name = L["Enable"],
				desc = L["Enable"],
				get = function()
					return Quartz:IsModuleActive('Buff')
				end,
				set = function(v)
					Quartz:ToggleModuleActive('Buff', v)
				end,
				order = 100,
			},
			focus = {
				type = 'group',
				name = L["Focus"],
				desc = L["Focus"],
				order = 101,
				args = {
					show = {
						type = 'toggle',
						name = L["Enable Focus"],
						desc = L["Enable Focus"],
						get = get,
						set = set,
						passValue = 'focus',
						order = 99,
					},
					buffs = {
						type = 'toggle',
						name = L["Enable Buffs"],
						desc = L["Enable Buffs"],
						get = get,
						set = set,
						passValue = 'focusbuffs',
					},
					debuffs = {
						type = 'toggle',
						name = L["Enable Debuffs"],
						desc = L["Enable Debuffs"],
						get = get,
						set = set,
						passValue = 'focusdebuffs',
					},
					position = {
						type = 'text',
						name = L["Position"],
						desc = L["Position"],
						get = get,
						set = set,
						validate = positions,
						passValue = 'focusposition',
					},
					showicons = {
						type = 'toggle',
						name = L["Show Icon"],
						desc = L["Show Icon"],
						get = get,
						set = set,
						passValue = 'focusicons',
						order = 101,
					},
					iconside = {
						type = 'text',
						name = L["Icon Side"],
						desc = L["Icon Side"],
						get = get,
						set = set,
						validate = {L["Left"], L["Right"]},
						passValue = 'focusiconside',
						order = 101,
					},
				},
			},
			target = {
				type = 'group',
				name = L["Target"],
				desc = L["Target"],
				order = 102,
				args = {
					show = {
						type = 'toggle',
						name = L["Enable Target"],
						desc = L["Enable Target"],
						get = get,
						set = set,
						passValue = 'target',
						order = 99,
					},
					buffs = {
						type = 'toggle',
						name = L["Enable Buffs"],
						desc = L["Enable Buffs"],
						get = get,
						set = set,
						passValue = 'targetbuffs',
					},
					debuffs = {
						type = 'toggle',
						name = L["Enable Debuffs"],
						desc = L["Enable Debuffs"],
						get = get,
						set = set,
						passValue = 'targetdebuffs',
					},
					position = {
						type = 'text',
						name = L["Position"],
						desc = L["Position"],
						get = get,
						set = set,
						validate = positions,
						passValue = 'targetposition',
					},
					showicons = {
						type = 'toggle',
						name = L["Show Icon"],
						desc = L["Show Icon"],
						get = get,
						set = set,
						passValue = 'targeticons',
						order = 101,
					},
					iconside = {
						type = 'text',
						name = L["Icon Side"],
						desc = L["Icon Side"],
						get = get,
						set = set,
						validate = {L["Left"], L["Right"]},
						passValue = 'targeticonside',
						order = 101,
					},
				},
			},
			bufftexture = {
				type = 'text',
				name = L["Texture"],
				desc = L["Set the Bar Texture"],
				validate = sm:List('statusbar'),
				order = 103,
				get = get,
				set = set,
				passValue = 'bufftexture',
			},
			buffwidth = {
				type = 'range',
				name = L["Buff Bar Width"],
				desc = L["Buff Bar Width"],
				min = 50,
				max = 300,
				step = 1,
				get = get,
				set = set,
				passValue = 'buffwidth',
				order = 104,
			},
			buffheight = {
				type = 'range',
				name = L["Buff Bar Height"],
				desc = L["Buff Bar Height"],
				min = 4,
				max = 25,
				step = 1,
				get = get,
				set = set,
				passValue = 'buffheight',
				order = 105,
			},
			buffnametext = {
				type = 'toggle',
				name = L["Buff Name Text"],
				desc = L["Buff Name Text"],
				get = get,
				set = set,
				passValue = 'buffnametext',
				order = 106,
			},
			bufftimetext = {
				type = 'toggle',
				name = L["Buff Time Text"],
				desc = L["Buff Time Text"],
				get = get,
				set = set,
				passValue = 'bufftimetext',
				order = 107,
			},
			bufffont = {
				type = 'text',
				name = L["Font"],
				desc = L["Font"],
				validate = sm:List('font'),
				get = get,
				set = set,
				passValue = 'bufffont',
				order = 108,
			},
			bufffontsize = {
				type = 'range',
				name = L["Font Size"],
				desc = L["Font Size"],
				min = 3,
				max = 15,
				step = 1,
				get = get,
				set = set,
				passValue = 'bufffontsize',
				order = 109,
			},
			buffalpha = {
				type = 'range',
				name = L["Alpha"],
				desc = L["Alpha"],
				min = 0.05,
				max = 1,
				step = 0.05,
				isPercent = true,
				get = get,
				set = set,
				passValue = 'buffalpha',
				order = 110,
			},
			buffcolor = {
				type = 'color',
				name = L["Buff Color"],
				desc = L["Buff Color"],
				get = getcolor,
				set = setcolor,
				passValue = 'buffcolor',
				order = 129,
			},
			debuffcolor = {
				type = 'color',
				name = L["Debuff Color"],
				desc = L["Debuff Color"],
				get = getcolor,
				set = setcolor,
				passValue = 'debuffcolor',
				order = 130,
			},
			bufftextcolor = {
				type = 'color',
				name = L["Text Color"],
				desc = L["Text Color"],
				get = getcolor,
				set = setcolor,
				passValue = 'bufftextcolor',
				order = 131,
			},
		},
	}
end